///////////////////////////////////////////////////////////////////////
//
// Multiple language Template Expansion
//
// Author: Mark A. Swann - 818-363-4311 (CA home phone, leave message.)
//         swannm@littongcs.com
//
//
// Profile: tab_size = 4
//
//
//  
// Purpose:
//  
//  This eel extension:
//  
//    1.  Binds a template expansion facility for multiple languages
//        to the space bar.
//    2.  Privides code coloring for ini files.
//  
//  
//  
// Installation:
//  
//  1.  Copy the files gui.mnu and template.ini to the Epsilon root
//      directory.
//  2.  Copy templts.e and templts.b into the Epsilon source directory.
//  3.  Load templts.b to install template expansion support.
//  4.  Save your state by running the write-state command 
//      normally found on C-F-3.
//  
//  
//  
// Description:
//  
//  The following commands and mode control variables, described below,
//  are provided:
//  
//      command expand_template();
//      command edit_templates();
//      command normal_character();
//      
//      user char           template_expansion       = 1;
//      user short          template_delay           = 35;
//      user buffer char    keep_trailing_whitespace = 0;
//
//  Additionally, a new "Templates" menu is provided on the menu bar which
//  allows you to interactively insert a template from a list, edit the
//  template file or clear the template table.
//  
//  
//  
// expand_template():
// ------------------
//  
//  The space character is normally bound to maybe_break_line. This
//  eel extension replaces the key binding for maybe_break_line with
//  expand_template. With expand_template bound to the space character,
//  all you need to do is enter one or more characters of the desired code
//  from template.ini then press the space bar. If you press another key
//  quickly after pressing the space bar, template expansion will be aborted
//  with both the space and the second key being processed normally. If the
//  2nd key is also a space then maybe_break_line will only be called once
//  causing only 1 space to be entered.
//  
//  Template expansion will not happen, if (1) template expansion is disable
//  (template_expansion = 0), (2) the current buffer mode is not supported in
//  the template.ini file, or (3) the code entered prior to the space is not
//  in template.ini.  You may prevent template expansion by double striking
//  the space bar within the template_delay timeout.
//  
//  
//  
// edit_templates():
// -----------------
//  
//  The template file (template.ini) can be modified with the edit_templates
//  command. This command replaces the buffer in the current window with
//  template.ini. You may edit the settings then press Alt-F3 (compile-buffer)
//  to process and save the file. The changes will automatically be reloaded.
//  
//  If code coloring is enabled, the various fields will be colored according
//  to the color classes for ini files defined in this extension. Any fields
//  that are not recognized or are found to contain errors will be colored
//  with the color class ini_error (normally white on red). After pressing
//  Alt-F3 the file will be scanned for any characters colored with ini_error.
//  If any are found, an error message is displayed and the file is saved but
//  it is not loaded.
//  
//
//
// normal_character():
// -------------------
//
//  This extension replaces the normal_character command found in basic.e
//  with a new version. Buffer modes that support auto-indentation will remove
//  any trailing white-space when you presses the enter key. It may be useful
//  during template expansion to disable this feature.  A new buffer variable,
//  keep_trailing_whitespace, will facilitate this (see below).
//
//
//  
// template_delay:
// ---------------
//  
//  The delay provided after pressing the space bar can be adjusted by
//  setting the variable template_delay which, by default, is set to
//  35 hundredths of a second. If this is too long or too short then you
//  may adjust it accordingly. 
//  
//  
//  
// template_expansion:
// ------------------
//  
//  Setting template_expansion to 0 disables template expansion.
//
//
//
// keep_trailing_whitespace:
// -------------------------
//
//  Setting this variable to 1 will disable the default behavior of
//  normal_character to remove any trailing white-space (see above). This
//  variable may be temporarily modified during template expansion
//  according to the KEEP_TRAILING_WHITESPACE parameter in the template file
//  (see template.ini).
//
//
// 
// Change log:
//
// 15 February 2001 Revised for Epsilon 10 by Lugaru, so that comment
// auto-filling works, and so that only the template.ini file gets
// special coloring logic.


#include "eel.h"
#include "kill.h"
#include "c.h"
#include "colcode.h"
#include "proc.h"
#include "inimode.h"

#define TEMPLATE_MACROS

//
// Template list structures and typedefs
//

typedef struct {
	short flags;						// see below
	short len;							// length of codeword
	char* codeword;						// codeword used in template search
	char* desc;							// descriptive text for menu display
	char* text;							// actual template text
} t_template;

typedef struct {
	char* mode_name;					// Name of buffer mode
	char* pattern;						// Pattern to recognize this class of code words
	char* marker;						// Cursor position marker
} t_lang;

typedef union {
	t_lang lang;
	t_template tmplt;
} t_info ;


typedef struct {
	char  type	;						// type of template list entry (see below)
	t_info info;
} t_tmplt_list;

//
// List of templates
//

t_tmplt_list *templates_p;

//
// Template list entry types
//

#define TMPLT_END		(0)
#define TMPLT_TEMPLATE	(1)
#define TMPLT_MACRO		(2)
#define TMPLT_LANGUAGE	(3)

//
// Flags
//

#define TMPL_NO_SMART_INDENT	(0x0001) // no automatic reindenting
#define TMPL_ADD_WHITESPACE		(0x0002) // add whitespace before template if not present
#define TMPL_LAST_WORD			(0x0004) // expand only if last word on line
#define TMPL_FIRST_WORD			(0x0008) // expand only if first word on line
#define TMPL_ON_NEW_LINE		(0x0010) // open new line before expanding
#define TMPL_CLOSE_BACK			(0x0040) // braces not indented
#define TMPL_HAS_MARKER			(0x0080) // set if marker is in template text
#define TMPL_CASE_FOLD			(0x0100) // ignore case in code search
#define TMPL_HAS_CASE_FOLD		(0x0200) // set if IGNORE_CASE specified
#define TMPL_KEEP_TRAILING_WS	(0x0400) // keep trailing whitespace
#define TMPL_USE_TABS			(0x0800) // insert tabs when tabbing
#define TMPL_HAS_USE_TABS		(0x1000) // set if INDENT_WITH_TABS is specified

//
// Mode control
//

user		char	template_expansion			= 1;
user		short	template_delay				= 35;			// in hundredths of a second
user		int		comment_look_back			= 10000;
user buffer char	keep_trailing_whitespace	= 0;

char* template_mode;					// Template buffer mode for completion

t_tmplt_list *pNULL_TEMPLATE = 0;		// NULL pointer to template

zeroed			t_tmplt_list	*common_lang_p; // Template list "Common" language header address
zeroed buffer	t_tmplt_list	*mode_lang_p; // Template list buffer language header address
zeroed buffer	short			template_rev=0;	// Template list revision number

//
// forward references
//

template_msg_box ( char* msg );
char *tmplt_match (char *s, int flags);
int in_comment ( int lookback );
int create_template_macro ( char* name, char* text, int smart_indent, char* marker );
command maybe_break_line();

//
// Template.ini file color classes
//

color_scheme "standard-color" {
	color_class ini_punctuation yellow;
	color_class ini_section_id white;
	color_class ini_mnem MAKE_RGB(255, 128, 128);
	color_class ini_parm_name MAKE_RGB(255, 128, 64);
	color_class ini_templt_name green on black;
	color_class ini_value cyan;
	color_class ini_error white on red;
	color_class ini_macro_error white on red;
};

color_scheme "standard-mono" {
	color_class ini_punctuation dark_grey;
	color_class ini_section_id white;
	color_class ini_mnem light_grey;
	color_class ini_parm_name light_grey;
	color_class ini_templt_name grey on black;
	color_class ini_value white;
	color_class ini_error black on white;
	color_class ini_macro_error black on white;
};

color_scheme "standard-gui" {
	color_class ini_punctuation		blue;
	color_class ini_section_id		MAKE_RGB(64, 0, 64) on yellow;
	color_class ini_mnem			MAKE_RGB(64, 0, 64);
	color_class ini_parm_name		MAKE_RGB(255, 128, 0);
	color_class ini_templt_name		MAKE_RGB(255, 128, 0);
	color_class ini_value			MAKE_RGB(0, 128, 64);
	color_class ini_error			white on red;
	color_class ini_macro_error		white on red;
};

// Use these settings as defaults for all other color classes.
color_class ini_punctuation =	color_class c_punctuation;
color_class ini_section_id	=	color_class c_comment;
color_class ini_mnem		=	color_class c_function;
color_class ini_parm_name	=	color_class c_identifier;
color_class ini_templt_name	=	color_class c_identifier;
color_class ini_value		=	color_class c_string;
color_class ini_error			white on red;
color_class ini_macro_error		white on red;

//
// Template file name
//

char template_file[] = "template.ini";

//
// Template file search pattern
//

#define TEMPLATE_PATTERN "^<Space|Tab>*(%[(<^Rbracket>+)%].*|\
parm<Space|Tab>+(<^Space|Tab|=>+)<Space|Tab>*=<Space|Tab>*(<^Space|Tab|Newline>.*)|\
tmplt<^Space|Tab>*<Space|Tab>+(<^Tab>+)<Tab>+(code)<Space|Tab>*%((<^Newline|Rparen>+)%)<Space|Tab>*=<Space|Tab>*(<^Space|Tab|Newline>.*))$"

//
// Pattern groups:
// 
// 0 = whole thing
// 1 = whole thing after leading ws
// 2 = section id
//		"parm"
// 3 = parm name
//		'='
// 4 = parm value
//		"tmplt"
// 5 = template descriptive label
// 6 = "code"
// 7 = template code
//		'='
// 8 = template value
// 

//
// Group names
//

#define SECTION_ID			(2)
#define PARM_NAME			(3)
#define PARM_VALUE			(4)
#define TEMPLATE_DESC		(5)
#define TEMPLATE_CODE_MNEM	(6)
#define TEMPLATE_CODE		(7)
#define TEMPLATE_VALUE		(8)

//
// Template file parameter keyword definitions
//

#define PARM_INVALID			(0)
#define PARM_CODE_PATTERN		(1)
#define PARM_ADD_WHITESPACE		(2)
#define PARM_FIRST_WORD			(3)
#define PARM_LAST_WORD			(4)
#define PARM_CLOSE_BACK			(5)
#define PARM_ON_NEW_LINE		(6)
#define PARM_SMART_INDENT		(7)
#define PARM_MARKER				(8)
#define PARM_IGNORE_CASE		(9)
#define PARM_KEEP_WHITESPACE	(10)
#define PARM_INDENT_WITH_TABS	(11)

////////////////////////////////////////////////////////////////////////////////
//
// Section: Template file initialization, editing, compiling
//
////////////////////////////////////////////////////////////////////////////////

//
// Interpret template file parameter keyword
//

int find_tmplt_parm ( char* parm )
{
	if ( strcmp ( parm, "CODE_PATTERN" ) == 0 ) {
		return PARM_CODE_PATTERN;
	}
	if ( strcmp ( parm, "ADD_WHITESPACE" ) == 0 ) {
		return PARM_ADD_WHITESPACE;
	}
	if ( strcmp ( parm, "FIRST_WORD") == 0 ) {
		return PARM_FIRST_WORD;
	}
	if ( strcmp ( parm, "LAST_WORD") == 0 ) {
		return PARM_LAST_WORD;
	}
	if ( strcmp ( parm, "CLOSE_BACK") == 0 ) {
		return PARM_CLOSE_BACK;
	}
	if ( strcmp ( parm, "ON_NEW_LINE") == 0 ) {
		return PARM_ON_NEW_LINE;
	}
	if ( strcmp ( parm, "SMART_INDENT" ) == 0 ) {
		return PARM_SMART_INDENT;
	}
	if ( strcmp ( parm, "MARKER" ) == 0 ) {
		return PARM_MARKER;
	}
	if ( strcmp ( parm, "IGNORE_CASE" ) == 0 ) {
		return PARM_IGNORE_CASE;
	}
	if ( strcmp ( parm, "KEEP_TRAILING_WHITESPACE" ) == 0 ) {
		return PARM_KEEP_WHITESPACE;
	}
	if ( strcmp ( parm, "INDENT_WITH_TABS" ) == 0 ) {
		return PARM_INDENT_WITH_TABS;
	}
	return PARM_INVALID;
}


//
// Clear template table
//

command clear_templates ()
{
	t_tmplt_list *tmplt_p;
	if ( templates_p ) {
		tmplt_p = templates_p;
		while ( 1 ) {
			switch ( tmplt_p->type ) {
				case TMPLT_END:
					free ( templates_p );
					templates_p = NULL;
					return;

				case TMPLT_TEMPLATE:
					if ( tmplt_p->info.tmplt.codeword ) {
						free ( tmplt_p->info.tmplt.codeword );
					}
					if ( tmplt_p->info.tmplt.text ) {
						free ( tmplt_p->info.tmplt.text );
					}
					if ( tmplt_p->info.tmplt.desc ) {
						free ( tmplt_p->info.tmplt.desc );
					}
					break;
				case TMPLT_MACRO:
					if ( tmplt_p->info.tmplt.codeword ) {
						free ( tmplt_p->info.tmplt.codeword );
					}
					if ( tmplt_p->info.tmplt.text ) {
						drop_name ( tmplt_p->info.tmplt.text );
					}
					if ( tmplt_p->info.tmplt.desc ) {
						free ( tmplt_p->info.tmplt.desc );
					}
					break;

				case TMPLT_LANGUAGE:
					if ( tmplt_p->info.lang.mode_name) {
						free ( tmplt_p->info.lang.mode_name );
					}
					if ( tmplt_p->info.lang.pattern ) {
						free ( tmplt_p->info.lang.pattern );
					}
					if ( tmplt_p->info.lang.marker ) {
						free ( tmplt_p->info.lang.marker );
					}
					break;
			}
			tmplt_p++;
		}
	}
}

//
// Template table initialization
//

int init_templates ()
{
	t_tmplt_list *lang_p = NULL;
	int has_error = 0;
	int indx, temp_buf=0;
	int template_count;
	short flags = Closeback ? TMPL_CLOSE_BACK : 0;
	char* path;
	char* pBfr;
	char* pTmpltText;
	char bfr[120];
	char value_char;
	save_var bufnum;
	save_var window_number;

	flags |= case_fold.default ? TMPL_CASE_FOLD : 0;

	clear_templates ();

	path = lookpath(template_file);
	if ( !path ) {
		error ( "Can't find file %s", template_file );
	}

	if (!look_file(path)) {
		temp_buf = bufnum = tmp_buf();
		if ( do_file_read ( path, default_translation_type) ) {
			error ( "Can't find file %s", template_file );
		}
		*filename = 0;
	}

	save_var narrow_start = 0;
	save_var narrow_end = 0;
	save_var point = 0;
	save_var case_fold = 1;

	template_count = 0;
	while (re_search(1, TEMPLATE_PATTERN ) ) {
		template_count++;
	}

	if ( template_count == 0 ) {
		return 0;
	}

	templates_p = (t_tmplt_list*)malloc ( (template_count+1) * sizeof (t_tmplt_list) );

	if ( !templates_p ) {
		error ( "Template array allocation failure" );
	}

	point = 0;
	indx = 0;
	while (re_search(1, TEMPLATE_PATTERN ) ) {
		// note ( "%s", bfr );
		switch ( toupper( character ( find_group (1, 1) )) ) {
			case '[':
				templates_p[indx].type = TMPLT_LANGUAGE;
				grab ( find_group(SECTION_ID, 1), find_group(SECTION_ID, 0), bfr );
				// note ( "%s", bfr );
				templates_p[indx].info.lang.mode_name = strsave ( bfr );
				lang_p = &templates_p[indx];
				if ( strfcmp ( lang_p->info.lang.mode_name, "common" ) == 0 ) {
					common_lang_p = lang_p;
				}
				flags = 0;
				indx++;
				continue;
			case 'T':
				if ( !lang_p ) {
					continue;
				}
				grab ( find_group(1, 1), find_group(TEMPLATE_DESC, 1), bfr );
				if ( (pBfr=index ( bfr, ' ' ))!= 0 ) {
					*pBfr = 0;
				} else if ( (pBfr=index ( bfr, '\t' )) != 0 ) {
					*pBfr = 0;
				} else {
					continue;
				}
				templates_p[indx].info.tmplt.flags = flags;
				if ( strcmp ( bfr, "tmplt" ) == 0 ) {
					templates_p[indx].type = TMPLT_TEMPLATE;
				} else {
					templates_p[indx].type = TMPLT_MACRO;
				}
				pTmpltText = malloc ( find_group(TEMPLATE_VALUE, 0) - find_group(TEMPLATE_VALUE, 1) + 2 );
				if ( !pTmpltText ) {
					continue;
				}
				grab ( find_group(TEMPLATE_VALUE, 1), find_group(TEMPLATE_VALUE, 0), pTmpltText );

				if ( strstr ( pTmpltText, lang_p->info.lang.marker ? lang_p->info.lang.marker : "@" ) ) {
					templates_p[indx].info.tmplt.flags |= TMPL_HAS_MARKER;
				}
				// replace all "\n" with '\n'
				pBfr = pTmpltText;
				while ( (pBfr = strstr ( pBfr, "\\n" ) ) ) {
					*pBfr = '\n';
					strcpy ( pBfr+1, pBfr+2 );
				}
				// replace all "\t" with '\t'
				pBfr = pTmpltText;
				while ( (pBfr = strstr ( pBfr, "\\t" ) ) ) {
					*pBfr = '\t';
					strcpy ( pBfr+1, pBfr+2 );
				}
				// replace all "\b" with '\b'
				pBfr = pTmpltText;
				while ( (pBfr = strstr ( pBfr, "\\b" ) ) ) {
					*pBfr = '\b';
					strcpy ( pBfr+1, pBfr+2 );
				}

				if ( templates_p[indx].type == TMPLT_MACRO ) {
					if ( !create_template_macro (
												bfr,
												pTmpltText,
												!(templates_p[indx].info.tmplt.flags & TMPL_NO_SMART_INDENT),
												(templates_p[indx].info.tmplt.flags & TMPL_HAS_MARKER) ? (lang_p->info.lang.marker ? lang_p->info.lang.marker : "@") : NULL
											   )
					   )
					{
						free ( pTmpltText );
						has_error = 1;
						buffer_printf (
									   "-template-errors",
									   "Buffer %s %d: Macro error\n",
									   bufname,
									   lines_between ( 0, point ) + 1
									  );
						goto Exit;
					}
					free ( pTmpltText );
					templates_p[indx].info.tmplt.text = strsave ( bfr );
				} else {
					if ( !create_template_macro (
												"template-dummy",
												pTmpltText,
												!(templates_p[indx].info.tmplt.flags & TMPL_NO_SMART_INDENT),
												(templates_p[indx].info.tmplt.flags & TMPL_HAS_MARKER) ? (lang_p->info.lang.marker ? lang_p->info.lang.marker : "@") : NULL
											   )
					   )
					{
						has_error = 1;
						buffer_printf (
									   "-template-errors",
									   "Buffer %s %d: Macro error\n",
									   bufname,
									   lines_between ( 0, point ) + 1
									  );
						goto Exit;
					}
					drop_name ( "template-dummy" );
					templates_p[indx].info.tmplt.text = pTmpltText;
				}

				grab ( find_group(TEMPLATE_DESC, 1), find_group(TEMPLATE_DESC, 0), bfr );
				// note ( "%s", bfr );
				templates_p[indx].info.tmplt.desc = strsave ( bfr );
				grab ( find_group(TEMPLATE_CODE, 1), find_group(TEMPLATE_CODE, 0), bfr );
				// note ( "%s", bfr );
				templates_p[indx].info.tmplt.len = strlen ( bfr );
				templates_p[indx].info.tmplt.codeword = strsave ( bfr );
				break;
			case 'P':
				grab ( find_group(PARM_NAME, 1), find_group(PARM_NAME, 0), bfr );
				value_char = toupper ( character ( find_group(PARM_VALUE, 1) ) );
				switch ( find_tmplt_parm ( bfr ) ) {
					case PARM_MARKER:
						if ( !lang_p ) {
							continue;
						}
						if ( lang_p->info.lang.marker ) {
							free ( lang_p->info.lang.marker );
						}
						grab ( find_group(PARM_VALUE, 1), find_group(PARM_VALUE, 0), bfr );
						lang_p->info.lang.marker = strsave ( bfr );
						continue;
					case PARM_CODE_PATTERN:
						if ( !lang_p ) {
							continue;
						}
						if ( lang_p->info.lang.pattern ) {
							free ( lang_p->info.lang.pattern );
						}
						grab ( find_group(PARM_VALUE, 1), find_group(PARM_VALUE, 0), bfr );
						lang_p->info.lang.pattern = strsave ( bfr );
						continue;
					case PARM_ADD_WHITESPACE:
						if ( value_char == 'Y' ) {
							flags |= TMPL_ADD_WHITESPACE;
						} else {
							flags &= ~TMPL_ADD_WHITESPACE;
						}
						continue;
					case PARM_FIRST_WORD:
						if ( value_char == 'Y' ) {
							flags |= TMPL_FIRST_WORD;
						} else {
							flags &= ~TMPL_FIRST_WORD;
						}
						continue;
					case PARM_LAST_WORD:
						if ( value_char == 'Y' ) {
							flags |= TMPL_LAST_WORD;
						} else {
							flags &= ~TMPL_LAST_WORD;
						}
						continue;
					case PARM_CLOSE_BACK:
						if ( value_char == 'Y' ) {
							flags |= TMPL_CLOSE_BACK;
						} else {
							flags &= ~TMPL_CLOSE_BACK;
						}
						continue;
					case PARM_IGNORE_CASE:
						flags |= TMPL_HAS_CASE_FOLD;
						switch ( value_char ) {
							case 'N':
								flags &= ~TMPL_CASE_FOLD;
								break;
							case 'Y':
								flags |= TMPL_CASE_FOLD;
								break;
							case 'D':
							default:
								flags &= ~TMPL_HAS_CASE_FOLD;
								break;
						}
						continue;
					case PARM_INDENT_WITH_TABS:
						flags |= TMPL_HAS_USE_TABS;
						switch ( value_char ) {
							case 'N':
								flags &= ~TMPL_USE_TABS;
								break;
							case 'Y':
								flags |= TMPL_USE_TABS;
								break;
							case 'D':
							default:
								flags &= ~TMPL_HAS_USE_TABS;
								break;
						}
						continue;
					case PARM_ON_NEW_LINE:
						if ( value_char == 'Y' ) {
							flags |= TMPL_ON_NEW_LINE;
						} else {
							flags &= ~TMPL_ON_NEW_LINE;
						}
						continue;
					case PARM_KEEP_WHITESPACE:
						if ( value_char == 'Y' ) {
							flags |= TMPL_KEEP_TRAILING_WS;
						} else {
							flags &= ~TMPL_KEEP_TRAILING_WS;
						}
						continue;
					case PARM_SMART_INDENT:
						if ( value_char == 'Y' ) {
							flags &= ~TMPL_NO_SMART_INDENT;
						} else {
							flags |= TMPL_NO_SMART_INDENT;
						}
						continue;
				}
			default:
				continue;
		}
		indx++;
		if ( indx >= template_count ) {
			break;
		}
	}

Exit:

	restore_vars();

	if ( temp_buf ) {
		buf_delete ( temp_buf );
	}
	template_rev.default++;
	return has_error;
}

int compile_templates ()
{
	int p, color, start, end;
	if ( modified ) {
		save_file ();
	}

	zap ( "-template-errors" );

	init_templates ();

	for (p=0; p < size(); ) {
		color = get_character_color(p, &start, &end);
		if (color == color_class ini_error ) {
			buffer_printf (
						   "-template-errors",
						   "Buffer %s %d: Not recognized\n",
						   bufname,
						   lines_between ( 0, p ) + 1
						  );
		} 
#if 0
else if (color == color_class ini_macro_error ) {
			buffer_printf (
						   "-template-errors",
						   "Buffer %s %d: Macro error\n",
						   bufname,
						   lines_between ( 0, p ) + 1
						  );
		}
#endif
		p = end;
	}

	if ( buffer_size ( "-template-errors" ) ) {
		when_setting_want_code_coloring();
		if (is_gui && want_gui_prompts) {
			template_msg_box ( "Template file contains errors.\nNot loaded." );
		} else {
			say ( "Template file contains errors. Not loaded." );
		}
		restore_vars();
		save_var bufname = "-template-errors";
		if (size()) {
			point = 0;			// Starting at the beginning,
			set_error_spot();	//    scan in this buffer.
			do_next_error(1);	// Scan for error messages.
		} else {
			restore_vars();
			delete_buffer("-template-errors");
		}
		return 1;
	} else {
		delete_buffer("-template-errors");
	}

	when_setting_want_code_coloring();
	say("Templates loaded.");
	return 1;
}

//
// Edit Template List
//

command edit_templates ()
{
	char* path;


	path = lookpath(template_file);
	if ( !path ) {
		error ( "Can't find file %s", template_file );
	}

	if ( look_file(path) ) {
		to_buffer(bufname);
	} else {
		zap ( template_file );
		to_buffer ( template_file );

		if ( do_file_read ( path, default_translation_type) ) {
			delete_buffer ( template_file );
			error ( "Can't find file %s", template_file );
		}
	}

	ini_mode ();

	say ( "Edit the templates list, then type Alt-F3 to compile, save and load the changes.");
}

////////////////////////////////////////////////////////////////////////////////
//
// Section: Template expansion
//
////////////////////////////////////////////////////////////////////////////////

//
// Dummy template_macro
//

command template_macro ()
{
}

void _purge_marker_ ()
{
	char tmp[2*FNAMELEN];

	get_string ( tmp, "" );

	if ( search ( -1, tmp ) ) {
		delete ( matchend, matchstart );
	}
}

int create_template_macro ( char* name, char* text, int smart_indent, char* marker )
{
	short *keyseq;
	int tmp;

	tmp = tmp_buf();
	save_var bufnum = tmp;
	insert ( '"' );
	stuff ( text );
	if ( smart_indent ) {
		stuff ( "<!do-reindent-after-yank>" );
	}
	if ( marker ) {
		stuff ( "<!_purge_marker_>" );
		stuff ( marker );
		stuff ( "C-M" );
	}
	insert ( '"' );
	point = 0;

	keyseq = get_keycode();
	restore_vars();
	buf_delete ( tmp );
	if ( !keyseq ) {
		return 0;
	}
	name_macro ( name, keyseq );
	return 1;
}

//
// Convert 'text' to a macro key sequence then append it to 'macro_list'
//

append_macro_string ( char* text )
{
	int tmp;
	short *keyseq;		/* sequence of key codes, in format used by macros */
	short *keyseq_p;

	if (_len_def_mac > 0 && !recording_suspended) {
		tmp = tmp_buf();
		save_var bufnum = tmp;
		insert ( '"' );
		stuff ( text );
		insert ( '"' );
		point = 0;

		keyseq = get_keycode();
		if ( !keyseq ) {
			restore_vars();
			buf_delete ( tmp );
			return;
		}
		if ( (_len_def_mac+*keyseq) >= MAX_MACRO ) {
			restore_vars();
			buf_delete ( tmp );
			end_kbd_macro();
			error("Macro definition buffer full: "
				  "keyboard macro defined");
		}


		cmd_len = 0;

		keyseq_p = keyseq + 1;

		while (  (--keyseq[0]) > 0 ) {
			def_text[_len_def_mac++] = *keyseq_p;
			cmd_len++;
			if (_len_def_mac >= MAX_MACRO) {
				restore_vars();
				buf_delete ( tmp );
				end_kbd_macro();
				error("Macro definition buffer full: "
					  "keyboard macro defined");
			}
		}

		restore_vars();
		buf_delete ( tmp );
	}
}

//
// Insert template into buffer
//

insert_template ( t_tmplt_list* template, int open_new_line, char* marker )
{
	short reg_tab_save = reg_tab[' '];
	save_var reindent_after_yank = 10000;
	save_var mark = point;

	dt_get ();							// Get time/date for time/date macros

	//
	// Select style
	//

	save_var Closeback = (template->info.tmplt.flags & TMPL_CLOSE_BACK) ? 1 : 0;
	if ( template->info.tmplt.flags & TMPL_HAS_CASE_FOLD ) {
		save_var case_fold = (template->info.tmplt.flags & TMPL_CASE_FOLD)  ? 1 : 0;
	}
	if ( template->info.tmplt.flags & TMPL_HAS_USE_TABS ) {
		save_var indent_with_tabs = (template->info.tmplt.flags & TMPL_USE_TABS)  ? 1 : 0;
	}
	save_var keep_trailing_whitespace = (template->info.tmplt.flags & TMPL_KEEP_TRAILING_WS) ? 1 : 0;

	if ( open_new_line ) {
		insert ( '\n' );				// Open a new line
	}

	if ( template->type == TMPLT_TEMPLATE ) {
		if ( !create_template_macro (
									 "template-macro",
									 template->info.tmplt.text,
									 !(template->info.tmplt.flags & TMPL_NO_SMART_INDENT),
									 (template->info.tmplt.flags & TMPL_HAS_MARKER) ? (marker ? marker : "@") : NULL
									)
		   )
		{
			return;
		}
		save_var recording_suspended = 1;
		reg_tab[' '] = maybe_break_line;

		template_macro ();

		reg_tab[' '] = reg_tab_save;
	} else {
		save_var recording_suspended = 1;
		reg_tab[' '] = maybe_break_line;

		try_calling ( template->info.tmplt.text );

		reg_tab[' '] = reg_tab_save;
	}
}

//
// Scan template table for user entry match
//

t_tmplt_list* scan_language_group ( t_tmplt_list *p_list, char* lookfor, int len )
{
	save_var case_fold = 1;
	p_list++;							// Skip past language header
	if ( len == 0 ) {
		for ( ; p_list->type; p_list++ ) {
			if ( p_list->type == TMPLT_LANGUAGE) {
				break;
			}
			// note ( "lookfor = \"%s\", codeword = \"%s\"", lookfor, p_list->codeword );
			if ( strfcmp ( lookfor, p_list->info.tmplt.desc ) == 0 ) {
				return p_list;
			}
		}
	} else {
		for ( ; p_list->type; p_list++ ) {
			if ( p_list->type == TMPLT_LANGUAGE) {
				break;
			}
			if ( len > p_list->info.tmplt.len ) {
				continue;
			}
			// note ( "lookfor = \"%s\", codeword = \"%s\"", lookfor, p_list->codeword );
			if ( strnfcmp ( lookfor, p_list->info.tmplt.codeword, MIN ( p_list->info.tmplt.len, len ) ) == 0 ) {
				return p_list;
			}
		}
	}
	return pNULL_TEMPLATE;
}

//
// Expand template according to user entry.
//

command expand_template () on reg_tab[' ']
{
	TIMER ts;
	char buf [128], *gripe_text;
	char is_first_word, open_new_line;
	int len;
	int first_key, save_key;
	int code_start;
	t_tmplt_list *tmplt, *got_it;
	
	save_var case_fold = 1;

	//
	// Exit if feature not enabled
	//

	if ( !template_expansion ) {
		maybe_insert_space ();
		return;
	}
		
	//
	// If user desires multiple insertions, don't expand.
	//

	if ( has_arg ) {
		maybe_insert_space ();
		return;
	}

	//
	// Initialize template table if needed
	//

	if ( !templates_p ) {
		init_templates ();
	}

	if ( !templates_p ) {				// No template list
		maybe_insert_space ();			// Insert space, maybe
		return;							// Exit
	}

	//
	// Start the user abort timer
	//

	if ( !run_by_mouse && !in_macro () && !key_from_macro ) {
		time_begin ( &ts, template_delay );
	}

	//
	// Scan template table for a matching buffer mode
	//

	if ( !mode_lang_p || !template_rev || template_rev != template_rev.default ) {
		mode_lang_p = NULL;
		for ( tmplt=templates_p; tmplt->type; tmplt++ ) {
			if ( tmplt->type == TMPLT_LANGUAGE && strcmp ( tmplt->info.lang.mode_name, major_mode ) == 0 ) {
				mode_lang_p = tmplt;
				template_rev = template_rev.default;
				break;
			}
		}
	}
	if ( !mode_lang_p && !common_lang_p ) {
		if ( run_by_mouse ) {
			template_msg_box ( "Template expansion not supported for the current buffer mode." );
		}
		maybe_insert_space ();			// Insert space
		return;							// no matching mode found
	}

	save_var narrow_start = 0;
	save_var narrow_end = 0;

	got_it = NULL;

	if ( run_by_mouse ) {
		template_mode = major_mode;
		*buf = 0;
		comp_read ( buf, "Select template to insert: ", tmplt_match, MUST_MATCH, "" );
		if ( !*buf ) {
			return;
		}
		//
		// Scan template table for user entry match
		//
		if ( mode_lang_p ) {
			got_it = scan_language_group ( mode_lang_p, buf, 0 );
		}
		if ( !got_it && common_lang_p ) {
			got_it = scan_language_group ( common_lang_p, buf, 0 );
		}
		save_var mark = point;				// Save mark. Set it to point.
		if ( got_it ) {
			stuff ( got_it->info.tmplt.codeword );
			exchange_point_and_mark ();
		}
	} else {
		save_var mark = point;				// Save mark. Set it to point.
		//
		// Check for codeword pattern match, scan template table for user entry match
		//
		if ( mode_lang_p && (len=parse_string ( -1, mode_lang_p->info.lang.pattern ? mode_lang_p->info.lang.pattern : "^<Tab|Space>*[a-z][a-z0-9_-]*", buf )) ) {
			got_it = scan_language_group ( mode_lang_p, buf, len );
		}
		if ( !got_it && common_lang_p ) {
			point = mark;
			if ( (len=parse_string ( -1, common_lang_p->info.lang.pattern ? common_lang_p->info.lang.pattern : "^<Tab|Space>*[a-z][a-z0-9_-]*", buf )) ) {
				got_it = scan_language_group ( common_lang_p, buf, len );
			}
		}
		if ( !got_it ) {
			point = mark;
			maybe_insert_space ();		// Insert space
			return;						//no match
		}
	}

	code_start = point;

	if ( got_it ) {
		if ( in_comment (comment_look_back) ) {
			gripe_text = "Expansion aborted (inside a comment).";
			goto maybe_gripe;
		}
		is_first_word = 0;
		point++;
		if ( parse_string ( -1, "^<Space|Tab>*<^Space|Tab>", NULL ) ) {
			is_first_word = 1;
		}
		point--;

		open_new_line = !is_first_word && (got_it->info.tmplt.flags & TMPL_ON_NEW_LINE);

		if ( !open_new_line && !is_first_word && got_it->info.tmplt.flags & TMPL_FIRST_WORD ) {
			gripe_text = "Expansion aborted. The chosen template must be the first word in line.";
			goto maybe_gripe;
		}

		if ( got_it->info.tmplt.flags & TMPL_LAST_WORD ) {
			point = mark - 1;
			if ( !parse_string ( 1, "<^Space|Tab><Space|Tab>*$", NULL ) ) {
				point = code_start;
				gripe_text = "Expansion aborted. The chosen template must be the last word in the line.";
				goto maybe_gripe;
			}
			point = code_start;
		}

		if ( !run_by_mouse ) {

			//
			// Wait a moment for user to enter another key in case he
			// isn't really doing a template expansion.
			//

			if ( !in_macro () && !key_from_macro  ) {
				delay ( time_remaining (&ts), COND_KEY );
				if ( char_avail () ) {
					goto user_abort_expansion;
				}
			}
			if ( _len_def_mac > 0 && !recording_suspended ) {
				if ( !key_from_macro ) {
					mac_unsave ();
				}
				append_macro_string ( "<!expand-template>" );
			}
		}

		if ( !open_new_line && point != 0 && got_it->info.tmplt.flags & TMPL_ADD_WHITESPACE ) {

			//
			// adjust if no previous whitespace
			//

			switch ( character ( point - 1 ) ) {
				case ' ':
				case '\n':
				case '\t':
					break;
				default:
					insert ( ' ' );
					break;
			}
		}

		//
		// Remove user entry
		//

		delete ( point, mark );

		//
		// Insert template, position the cursor
		//

		insert_template ( got_it, open_new_line, mode_lang_p->info.lang.marker );

		return;
	}

	goto abort_expansion;				// No codeword match.

maybe_gripe:

	if ( run_by_mouse ) {
		delete ( point, mark );
		template_msg_box ( gripe_text );
	} else if ( char_avail () ) {
		goto user_abort_expansion;
	}

	//
	// No codeword match. Insert space and exit
	//

abort_expansion:

	point = mark;
	restore_vars();
	maybe_insert_space ();

	return;

	//
	// Abort template expansion
	//

user_abort_expansion:

	first_key = key;			// new key, save first key
	getkey ();					// get second key

	//
	// Check for 2nd strike of template-expand key
	//

	if ( key == first_key ) {	// template-expand key struck twice
		point = mark;
		restore_vars();
		maybe_insert_space ();	// insert template-expand key
		return;
	}

	//
	// Second key differs, insert first key then process second key.
	//

	save_key = key;				// save second key
	key = first_key;			// restore first key
	point = mark;
	restore_vars();
	maybe_insert_space ();		// insert first key
	key = save_key;				// fetch second key
	do_again ();				// process it
}

//
// Fake-out maybe_break_line while in macro
//

tmplt_maybe_break_line ()
{
	if ( key_from_macro && _len_def_mac > 0 && !recording_suspended ) {
		append_macro_string ( "<!tmplt-maybe-break-line>" );
	}
	key = ' ';
	maybe_break_line ();
}

//
// Conditionally insert space character
//

maybe_insert_space ()
{
	if ( _len_def_mac > 0 && !recording_suspended ) {
		mac_unsave ();
		append_macro_string ( "<!tmplt-maybe-break-line>" );
	}

	if ( !run_by_mouse && key == ' ' ) {
		maybe_break_line();
	}
}

//
// Determine if point is inside a comment (return 1 if yes, 0 if no)
//

int in_comment ( int lookback )
{
	int save_point = point;
	save_var narrow_start = point - lookback;
	save_var matchstart, matchend, point;

	if ( !re_search ( RE_REVERSE, comment_start ) ) {
		return 0;
	}
	if ( re_search ( RE_FORWARD, comment_pattern ) ) {
		if ( save_point >= matchstart && save_point < matchend ) {
			return 1;
		}
	}
	return 0;
}

//
// Completion mode template match finder
//

t_tmplt_list *tmplt_match_p = 0;
char show_common = 0;

char *tmplt_match(s, flags)
	char *s;
{
	t_tmplt_list *tmplt_p;
	int len = strlen ( s );
	int desc_len;

	//
	// Initialize template table if needed
	//

	if ( !templates_p ) {
		init_templates ();
	}

	if ( !templates_p ) {				// No template list
		return NULL;					// Exit
	}

	if ( (flags & STARTMATCH) || !tmplt_match_p ) {
		tmplt_match_p = templates_p;
		show_common = 0;

		//
		// Scan template table for a matching buffer mode
		//

		for ( ; tmplt_match_p->type; tmplt_match_p++ ) {
			if ( tmplt_match_p->type == TMPLT_LANGUAGE && strcmp ( tmplt_match_p->info.lang.mode_name, template_mode ) == 0 ) {
				tmplt_match_p++;
				break;
			}
		}

		if ( !tmplt_match_p->type ) {
			return NULL;					// no matching mode found
		}
	}

	//
	// Search for match
	//

	tmplt_p = tmplt_match_p;
	for ( ; tmplt_p->type; tmplt_p++ ) {
		if ( tmplt_p->type == TMPLT_LANGUAGE) {
			if ( show_common ) {
				show_common = 0;
				tmplt_match_p = NULL;
				return NULL;
			}
			if ( !common_lang_p ) {
				tmplt_match_p = NULL;
				return NULL;
			}
			show_common = 1;
			tmplt_match_p = common_lang_p + 1;
			tmplt_p = common_lang_p;
			continue;
		}
		desc_len = strlen ( s );
		if ( len > desc_len ) {
			continue;
		}
		if ( (flags & EXACTONLY) && len != desc_len ) {
			continue;
		}
		// note ( "buf = \"%s\", codeword = \"%s\"", buf, tmplt_p->codeword );
		if ( strncmp ( s, tmplt_p->info.tmplt.desc, len ) == 0 ) {
			tmplt_match_p = tmplt_p+1;
			return tmplt_p->info.tmplt.desc;
		}
	}
	return NULL;
}

////////////////////////////////////////////////////////////////////////////////
//
// Section: "ini" file mode support
//
////////////////////////////////////////////////////////////////////////////////

//
// Functions for coloring ini files
//

color_ini_from_here(safe)
{
	point = safe;
	to_begin_line();
	return point;
}

color_template_ini_range(from, to) // recolor just this section
{
	char bfr[120];
	int tmp, bufnum_save;
	int section_start;
	int section_end;
	int parm_name_start;
	int parm_name_end;
	int parm_value_start;
	int parm_value_end;
	int tmplt_lbl_start;
	int tmplt_lbl_end;
	int tmplt_code_start;
	int tmplt_code_end;
	int tmplt_code_mnem_end;
	int tmplt_value_start;
	int tmplt_value_end;
	if (from >= to)
		return to;
	save_var point, matchstart, matchend;

	point = from;
	while ( point < size() && point < to ) {
		if ( parse_string ( 1, TEMPLATE_PATTERN, NULL ) ) {

			from = point;
			nl_forward();
			set_character_color(from, point, color_class ini_mnem );

			section_start		= find_group(SECTION_ID,		1);
			section_end			= find_group(SECTION_ID,		0);

			parm_name_start		= find_group(PARM_NAME,			1);
			parm_name_end		= find_group(PARM_NAME,			0);

			parm_value_start	= find_group(PARM_VALUE,		1);
			parm_value_end		= find_group(PARM_VALUE,		0);

			tmplt_lbl_start		= find_group(TEMPLATE_DESC,		1);
			tmplt_lbl_end		= find_group(TEMPLATE_DESC,		0);

			tmplt_code_mnem_end	= find_group(TEMPLATE_CODE_MNEM,	0);

			tmplt_code_start		= find_group(TEMPLATE_CODE,		1);
			tmplt_code_end		= find_group(TEMPLATE_CODE,		0);

			tmplt_value_start	= find_group(TEMPLATE_VALUE,	1);
			tmplt_value_end		= find_group(TEMPLATE_VALUE,	0);

			switch ( toupper ( character (find_group(1, 1)) ) ) {
				case '[':
					set_character_color ( find_group(1, 1), find_group(1, 0), color_class ini_punctuation );
					set_character_color ( section_start, section_end, color_class ini_section_id );
					break;
				case 'T':
					set_character_color ( tmplt_lbl_start, tmplt_lbl_end, color_class ini_templt_name );
					set_character_color ( tmplt_code_mnem_end, tmplt_code_start, color_class ini_punctuation );
					set_character_color ( tmplt_code_start, tmplt_code_end, color_class ini_templt_name );
					set_character_color ( tmplt_code_end, tmplt_value_start, color_class ini_punctuation );
					if ( (tmplt_value_end-tmplt_value_start) >= sizeof (bfr) ) {
						set_character_color ( tmplt_value_start, tmplt_value_end, color_class ini_value );
						break;
					}
					grab ( tmplt_value_start, tmplt_value_end, bfr );
					bufnum_save = bufnum;
					tmp = tmp_buf();
					bufnum = tmp;
					insert ( '"' );
					stuff ( bfr );
					insert ( '"' );
					point = 0;

					if ( !get_keycode() ) {
						bufnum = bufnum_save;
						buf_delete ( tmp );
						set_character_color ( tmplt_value_start, tmplt_value_end, color_class ini_macro_error );
					} else {
						bufnum = bufnum_save;
						buf_delete ( tmp );
						set_character_color ( tmplt_value_start, tmplt_value_end, color_class ini_value );
					}
					break;
				case 'P':
					compile_buffer_cmd = "";			// dummy
					pre_compile_hook = compile_templates;

					grab ( parm_name_start, parm_name_end, bfr );
					switch ( find_tmplt_parm ( bfr ) ) {
						case PARM_CODE_PATTERN:
							set_character_color ( parm_name_start, parm_name_end, color_class ini_parm_name );
							set_character_color ( parm_name_end, parm_value_start, color_class ini_punctuation );
							grab ( parm_value_start, parm_value_end, bfr );
							if ( re_compile ( RE_REVERSE, bfr ) ) {
								set_character_color ( parm_value_start, parm_value_end, color_class ini_error );
							} else {
								set_character_color ( parm_value_start, parm_value_end, color_class ini_value );
							}
							break;
						case PARM_MARKER:
							set_character_color ( parm_name_start, parm_name_end, color_class ini_parm_name );
							set_character_color ( parm_name_end, parm_value_start, color_class ini_punctuation );
							set_character_color ( parm_value_start, parm_value_end, color_class ini_value );
							break;
						case PARM_INVALID:
							set_character_color ( parm_name_start, parm_name_end, color_class ini_error );
							set_character_color ( parm_name_end, parm_value_start, color_class ini_punctuation );
							set_character_color ( parm_value_start, parm_value_end, color_class ini_value );
							break;
						default:
							set_character_color ( parm_name_start, parm_name_end, color_class ini_parm_name );
							set_character_color ( parm_name_end, parm_value_start, color_class ini_punctuation );

							if ( index ( "YyNnDd", character ( parm_value_start ) ) ) {
								set_character_color ( parm_value_start, parm_value_end, color_class ini_value );
							} else {
								set_character_color ( parm_value_start, parm_value_end, color_class ini_error );
							}
							break;
					}
					break;
				default:
					break;
			}
		} else if ( parse_string ( 1, "^<Space|Tab>*(<^;|=|Newline|Space>+)<Space|Tab>*=<Space|Tab>*(<^Space|Tab|Newline>.*)$", NULL ) ) {
			from = point;
			nl_forward();

			set_character_color(from, point, color_class ini_punctuation );

			set_character_color ( find_group(1, 1), find_group(1, 0), color_class ini_parm_name );
			set_character_color ( find_group(2, 1), find_group(2, 0), color_class ini_value );
		} else if ( parse_string ( 1, "^<Space|Tab>*(;)(.*)$", NULL ) ) {
			from = point;
			nl_forward();
			set_character_color ( find_group(1, 1), find_group(1, 0), color_class ini_punctuation );
			set_character_color ( find_group(2, 1), find_group(2, 0), -1 );
		} else if ( parse_string ( 1, "(^<Newline>)", NULL ) ) {
			point++;
		} else {
			from = point;
			to_end_line ();
			set_character_color(from, point, color_class ini_error  );
			point++;
		}
	}
	return point;
}

//
// Colorization debug routines
//

#if 0
command debug_color_region()
{
	fix_region();
	set_character_color(point, mark, color_class default);
	point = color_template_ini_range(point, mark);
}

command debug_from_here()
{
	point = color_ini_from_here(point);
}
#endif

//
// INI file mode initialization
//

char _template_ini_mode_name[] = "INI";

command template_ini_mode()
{
	mode_keys = buf_tab;		/* use these keys */
	major_mode = _template_ini_mode_name;

	indenter = indenter.default;
	if (!indenter)
		indenter = prev_indenter;
	strcpy(comment_start, comment_start.default);
	strcpy(comment_pattern, comment_pattern.default);
	strcpy(comment_begin, comment_begin.default);
	strcpy(comment_end, comment_end.default);

	recolor_range = color_template_ini_range;	// set up coloring rules
	recolor_from_here = color_ini_from_here;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	if ( strcmp ( bufname, template_file ) == 0 ) {
		compile_buffer_cmd = "";			// dummy
		pre_compile_hook = compile_templates;
	}
	make_mode();
}

suffix_ini ()
{
	if (!strfcmp(get_tail(filename, 1), "template.ini"))
		template_ini_mode();
	else
		ini_mode();
}

// Now include the original definitions of some Epsilon 10 functions,
// in case a user loaded the v9 version of this file, which would have
// overriden them.

color_ini_range(from, to) // recolor just this section
{			// last colored region may go past to
	int t = -1;

	if (from >= to)
		return to;
	save_var point, matchstart, matchend;
	point = to;				// Color entire lines.
	nl_forward();
	to = point;
	set_character_color(from, to, -1);
	point = from;
	to_begin_line();
	for (; point < to; nl_forward()) {
		t = point;
		switch (curchar()) {
			case '[':
				set_character_color(t, give_end_line() + 1,
									color_class inifile_section);
				break;
			case '#':
			case ';':
				set_character_color(t, give_end_line() + 1,
									color_class inifile_comment);
				break;
			default:	// Some textual .sys files use REM for comments.
				if (parse_string(1, "[ \t]*REM[ \t\n]")) {
					set_character_color(t, give_end_line(),
										color_class inifile_comment);
				} else if (parse_string(1, "[^\n=]*="))
					set_character_color(t, matchend,
										color_class inifile_definition);
		}
	}
	return point;
}

command ini_mode()
{
	mode_default_settings();
	major_mode = _ini_mode_name;
	strcpy(comment_start, ";[ \t]*");
	strcpy(comment_pattern, ";.*$");
	strcpy(comment_begin, "; ");
	strcpy(comment_end, "");
	comment_column = 0;
	recolor_range = color_ini_range;	// set up coloring rules
	recolor_from_here = recolor_by_lines;
	if (want_code_coloring)		// maybe turn on coloring
		when_setting_want_code_coloring();
	try_calling("ini-mode-hook");
	make_mode();
}


////////////////////////////////////////////////////////////////////////////////
//
// Section: Misc. template insertion support
//
////////////////////////////////////////////////////////////////////////////////

//
// Display simple message box
//

template_msg_box ( char* msg )
{
	if (is_gui && want_gui_prompts) {
		button_dialog ("", msg, "Ok", "", "", 1);
		note ( "" );
	} else {
		say ( msg );
	}
}

//
// Time/Date support commands
//

struct time_info template_time;

char dwk[7][4] = {
	"Sun",
	"Mon",
	"Tue",
	"Wed",
	"Thu",
	"Fri",
	"Sat"
};

char mon[12][4] = {
	"Jan",
	"Feb",
	"Mar",
	"Apr",
	"May",
	"Jun",
	"Jul",
	"Aug",
	"Sep",
	"Oct",
	"Nov",
	"Dec",
};

dt_get ()
{
	time_and_day(&template_time);
}

//
// Insert date
//

command ins_date ()
{
	bprintf (
			 "%d/%d/%d",
			 template_time.month,
			 template_time.day,
			 template_time.year
			);
}

//
// Insert time
//

command ins_time ()
{
	bprintf (
			 "%2d:%02d",
			 template_time.hour,
			 template_time.minute
			);
}

command ins_h12 ()
{
	short hour = template_time.hour % 12;
	if ( hour == 0 ) {
		hour = 12;
	}
	bprintf (  "%2d", hour );
}

command ins_h24 ()
{
	bprintf (  "%2d", template_time.hour );
}

command ins_min ()
{
	bprintf ( "%02d", template_time.minute );
}

command ins_sec ()
{
	bprintf ( "%02d", template_time.second );
}

command ins_am_pm ()
{
	bprintf ( "%s", template_time.hour >= 12 ? "Pm" : "Am" );
}

command ins_year ()
{
	bprintf ( "%d", template_time.year );
}

command ins_month ()
{
	bprintf ( "%d", template_time.month );
}

command ins_day ()
{
	bprintf ( "%d", template_time.day );
}

command ins_dname ()
{
	bprintf ( "%s", dwk[template_time.day_of_week%7] );
}

command ins_year ()
{
	bprintf ( "%d", template_time.year );
}

command ins_month ()
{
	bprintf ( "%d", template_time.month );
}

command ins_mname ()
{
	bprintf ( "%s", mon[(template_time.month-1)%12] );
}

//
// File info support commands
//

command ins_fname ()
{
	stuff ( get_tail ( filename, 0 ) );
}

command ins_fdate ()
{
	get_file_date ();
	bprintf (
			 "%d/%d/%d",
			 file_info.month,
			 file_info.day,
			 file_info.year
			);
}

command ins_ftime ()
{
	short hour;
	get_file_date ();
	hour = file_info.hour % 12;
	if ( hour == 0 ) {
		hour = 12;
	}
	bprintf (
			 "%2d:%02d %s",
			 hour,
			 file_info.minute,
			 file_info.hour >= 12 ? "Pm" : "Am"
			);
}

command ins_ftime_24 ()
{
	get_file_date ();
	bprintf (
			 "%2d:%02d",
			 file_info.hour,
			 file_info.minute
			);
}

//
// Misc. cursor and text positioning support
//

command bol ()
{
	iter=0;				/* throw away any argument */
	cursor_to_column = -1;
	to_begin_line();
}

command eol ()
{
	iter=0;			/* throw away any argument */
	cursor_to_column = -1;
	to_end_line();
}

command to_col ()
{
	if ( has_arg && iter > 0 ) {
		to_column ( iter - 1 );
	}
	iter = 1;
}

command allign_under ()
{
	save_var mark = point;
	bol ();
	delete_horizontal_space ();
	indent_under ();
	point = mark;
}

////////////////////////////////////////////////////////////////////////////////
//
// Section: Miscellaneous buffer mode support
//
////////////////////////////////////////////////////////////////////////////////

//
// Basic
//

suffix_bas ()
{
	fundamental_mode();
	major_mode = "Basic";
	auto_indent = 1;
}

//
// DBase
//

suffix_prg ()
{
	fundamental_mode();
	major_mode = "prg";
	auto_indent = 1;
}

//
// Cobol
//

suffix_cob ()
{
	fundamental_mode();
	major_mode = "Cobol";
	auto_indent = 1;
}

////////////////////////////////////////////////////////////////////////////////
//
// Section: Epsilon behavior override
//
////////////////////////////////////////////////////////////////////////////////

//
// Modify normal end-of-line behavior
//

maybe_delete_horizontal_space()
{
	if ( !keep_trailing_whitespace ) {
		delete_horizontal_space();
	}
}

/* insert or overwrite according to mode */
command normal_character() on reg_tab[CTRL('J')]	/* more, see below */
{
	int i;

	check_abort();
	delete_if_highlighted();
	if (!call_maybe_break_line(3))
		return;
	if (key >= 512) return;
	if (IS_ALT_KEY(key))
		key = key | 128;			// turn on meta bit
	if (key == '\n') {
		if (over_mode) {
			if (!nl_forward())
				insert('\n');
		} else if (auto_indent && indenter) {
#if 1
			maybe_delete_horizontal_space (); // new code
#else
			delete(point, point - parse_string(-1, "[ \t]*")); // original code
#endif
			insert('\n');		/* delete trailing whitesp */
			(*indenter)();
		} else
			insert('\n');
	} else if (over_mode && point < size() && curchar() != '\n') {
		if (curchar() == '\t')
			i = display_width('\t', current_column());
		else
			i = 0;
		replace(point++, key);
		while (--i > 0)		/* insert i-1 spaces after pt */
			insert(' '), point--;
	} else
		insert(key);
}
